==> quantarhei-0.0.5.tar.gz: quantarhei-0.0.5/quantarhei/qm/liouvillespace/redfieldrates.py <==
# -*- coding: utf-8 -*-

import numpy

from ...core.implementations import implementation
from ...core.units import cm2int
from ...core.units import kB_intK

from ..hilbertspace.hamiltonian import Hamiltonian
from ..liouvillespace.systembathinteraction import SystemBathInteraction

"""
*******************************************************************************

      REDFIELD RATE MATRIX

*******************************************************************************
"""     
class RedfieldRateMatrix:
    
    def __init__(self,ham,sbi,initialize=True,cutoff_time=None):
        
        if not isinstance(ham,Hamiltonian):
            raise Exception("First argument must be a Hamiltonian")
            
        if not isinstance(sbi,SystemBathInteraction):
            raise Exception
            
        self._is_initialized = False            
        self._has_cutoff_time = False
        
        if cutoff_time is not None:
            self.cutoff_time = cutoff_time
            self._has_cutoff_time = True            
            
        self.Hamiltonian = ham
        self.sbi = sbi
        
        if initialize: 
            self._set_rates(ham,sbi)          
            self._is_initialized = True
                
                
    def _set_rates(self,ham,sbi):
        """ Reference implementation, completely in Python
        
        """
        
        # dimension of the Hamiltonian (includes excitons
        # with all multiplicities specified at its creation)
        Na = ham.data.shape[0]
     
        # number of components
        Nk = self.sbi.N  
        
        if Nk <= 0:
            raise Exception("No system bath intraction components present")
        
        # Eigen problem
        hD,SS = numpy.linalg.eigh(ham.data) 
        S1 = numpy.linalg.inv(SS)
        
        # component operators
        KI = self.sbi.KK
        
        #FIXME: This has to be made configurable
        # frequency cut-off
        freq_cutoff = 3000.0*cm2int
        #print("freq. cut-off",freq_cutoff)
        
        # temperature
        Temp = self.sbi.CC.get_correlation_function(0,0).T
        



        """ Might need some speed-up too"""
        
        # transform interaction operators
        for i in range(Nk):
            KI[i,:,:] = numpy.dot(S1,numpy.dot(KI[i,:,:],SS))
            
        #  Find all eigenfrequencies 
        Om = numpy.zeros((Na,Na))
        for a in range(0,Na):
            for b in range(0,Na):
                Om[a,b] = hD[a] - hD[b]
                
        # calculate values of the spectral density at frequencies
        cc = numpy.zeros((Nk,Na,Na),dtype=numpy.complex)
        
        # loop over components
        for k in range(Nk):

            # spectral density
            cf = self.sbi.CC.get_correlation_function(k,k)
            cf.convert_2_spectral_density()

            # Spectral density at all frequencies
            for i in range(Na):
                for j in range(Na):
                    if i != j:
                        if numpy.abs(Om[i,j]) > freq_cutoff:
                            cc[k,i,j] = 0.0
                        else:
                            if Om[j,i] < 0.0:
                                cc[k,i,j] = (cf.interp_data(Om[i,j])
                                *numpy.exp(Om[j,i]/(kB_intK*Temp)))
                            else:
                                cc[k,i,j] = cf.interp_data(Om[j,i])
                                

        """ To submit: 
                        Na
                        Nk
                        KI[Nk,Na,Na]
                        cc[Nk,Na,Na]
                        
             
            To return:
                        RR
                        
        """
        
        warning = []
        error = []
        rtol = 1.0e-6
        self.data = ssRedfieldRateMatrix(Na,Nk,KI,cc,rtol,warning,error)
        for w in error:
            print(w)                                    
        self._is_initialized = True
        
        

@implementation("redfield","ssRedfieldRateMatrix",
                at_runtime=True,
                fallback_local=True,
                always_local=True)
def ssRedfieldRateMatrix(Na,Nk,KI,cc,rtol,warning,error):
    
    # output relaxatio rate matrix
    RR = numpy.zeros((Na,Na),dtype=numpy.float)
    cc = 2.0*numpy.real(cc)
    
    # loop over components
    for k in range(Nk):
        
        # interaction operator
        KK = KI[k,:,:]

        for i in range(Na):
            for j in range(Na):
                if i != j:                                
                            
                    RR[i,j] += (cc[k,i,j]*KK[i,j]*KK[j,i])
                    
                    if RR[i,j] < 0.0:
                        warning.append("\n%i %i %r smaller than zero"
                                    % (i,j,RR[i,j]))
                        if numpy.abs(RR[i,j]) < rtol:
                            RR[i,j] = 0.0
                        else:
                            error.append(" ... significantly")
                            
                    RR[j,j] -= RR[i,j]
                    
    return RR

==> quantarhei-0.0.8.tar.gz: quantarhei-0.0.8/quantarhei/qm/liouvillespace/redfieldrates.py <==
# -*- coding: utf-8 -*-
"""
*******************************************************************************

      REDFIELD RATE MATRIX

*******************************************************************************
"""  

import numpy

from ...core.implementations import implementation
from ...core.units import cm2int
from ...core.units import kB_intK

from ..hilbertspace.hamiltonian import Hamiltonian
from ..liouvillespace.systembathinteraction import SystemBathInteraction

   
class RedfieldRateMatrix:
    """Redfield relaxation rate matrix
    
    Redfield population relaxation rate matrix is calculated from the 
    Hamiltonian and system-system bath interation. The bath
    correlation functions are Fourier transformed by FFT and the negative
    frequency part is calculated from the expected thermodynamic
    symmetry. 
    
    Parameters
    ----------
    
    ham : Hamiltonian
        Hamiltonian object 
        
    sbi : SystemBathInteraction
        SystemBathInteraction object
        
    initialize : bool (default True)
        If true, the rates will be calculated when the object is created
        
    cutoff_time : float
        If cutoff time is specified, the tensor is integrated only up to the
        cutoff time
    
    
    """
    
    def __init__(self, ham, sbi, initialize=True, cutoff_time=None):
        
        if not isinstance(ham,Hamiltonian):
            raise Exception("First argument must be a Hamiltonian")
            
        if not isinstance(sbi,SystemBathInteraction):
            raise Exception
            
        self._is_initialized = False            
        self._has_cutoff_time = False
        
        if cutoff_time is not None:
            self.cutoff_time = cutoff_time
            self._has_cutoff_time = True            
            
        self.Hamiltonian = ham
        self.sbi = sbi
        
        if initialize: 
            self._set_rates(ham,sbi)          
            self._is_initialized = True
                
                
    def _set_rates(self,ham,sbi):
        """ Reference implementation, completely in Python
        
        """
        
        # dimension of the Hamiltonian (includes excitons
        # with all multiplicities specified at its creation)
        Na = ham.data.shape[0]
     
        # number of components
        Nk = self.sbi.N  
        
        if Nk <= 0:
            raise Exception("No system bath intraction components present")
        
        # Eigen problem
        hD,SS = numpy.linalg.eigh(ham.data) 
        S1 = numpy.linalg.inv(SS)
        
        # component operators
        KI = self.sbi.KK.copy()
        
        #FIXME: This has to be made configurable
        # frequency cut-off
        freq_cutoff = 3000.0*cm2int
        #print("freq. cut-off",freq_cutoff)
        
        # temperature
        Temp = self.sbi.CC.get_correlation_function(0,0).temperature
        
        
        # transform interaction operators
        for i in range(Nk):
            KI[i,:,:] = numpy.dot(S1,numpy.dot(KI[i,:,:],SS))
            
        #  Find all eigenfrequencies 
        Om = numpy.zeros((Na,Na))
        for a in range(0,Na):
            for b in range(0,Na):
                Om[a,b] = hD[a] - hD[b]
                
        # calculate values of the spectral density at frequencies
        cc = numpy.zeros((Nk,Na,Na),dtype=numpy.complex)
        
        # loop over components
        for k in range(Nk):

            # correlation function
            cf = self.sbi.CC.get_correlation_function(k,k)
            
            # Ft correlation function
            cw = cf.get_Fourier_transform()
            

            # Spectral density at all frequencies
            for i in range(Na):
                for j in range(Na):
                    if i != j:
                        if numpy.abs(Om[j,i]) > freq_cutoff:
                            cc[k,i,j] = 0.0
                        else:
                            if Om[j,i] < 0.0:
                                #cc[k,i,j] = (cf.interp_data(Om[i,j])
                                cc[k,i,j] = (cw.at(Om[i,j],approx="spline")
                                *numpy.exp(-Om[i,j]/(kB_intK*Temp)))
                            else:
                                cc[k,i,j] = cw.at(Om[j,i],approx="spline") 
                                

        """ To submit: 
                        Na
                        Nk
                        KI[Nk,Na,Na]
                        cc[Nk,Na,Na]
                        
             
            To return:
                        RR
                        
        """
        
        warning = []
        error = []
        rtol = 1.0e-6
        self.data = ssRedfieldRateMatrix(Na,Nk,KI,cc,rtol,warning,error)
        for w in error:
            print(w)     
                               
        self._is_initialized = True
        
        

@implementation("redfield","ssRedfieldRateMatrix",
                at_runtime=True,
                fallback_local=True,
                always_local=True)
def ssRedfieldRateMatrix(Na,Nk,KI,cc,rtol,warning,error):
    
    # output relaxatio rate matrix
    RR = numpy.zeros((Na,Na),dtype=numpy.float)
    
    # real part of FT correlation function = 2Re of the half FT of 
    # the correlation function - no factor of 2 here!
    cr = numpy.real(cc)
    
    # loop over components
    for k in range(Nk):
        
        # interaction operator
        KK = KI[k,:,:]

        for i in range(Na):
            for j in range(Na):
                
                # calculate rates, i.e. off diagonal elements
                if i != j:                                
                            
                    RR[i,j] += (cr[k,i,j]*KK[i,j]*KK[j,i])
                    
                    if RR[i,j] < 0.0:
                        warning.append("\n%i %i %r smaller than zero"
                                    % (i,j,RR[i,j]))
                        if numpy.abs(RR[i,j]) < rtol:
                            RR[i,j] = 0.0
                        else:
                            error.append(" ... significantly")
    # calculate the diagonal elements (the depopulation rates)            
    for i in range(Na):
        for j in range(Na):
            if i != j:
                RR[j,j] -= RR[i,j]
                    
    return RR
